from scipy.sparse import csr_matrix
from scipy.sparse import dok_matrix
from scipy.sparse import issparse
import scipy.sparse as spr
import numpy as np
from qutip import *


def PermMat(A):
    """
    Takes a matrix as input and returns
    a permutation matrix that makes it block 
    diagonal.

    Algorithm is linear in number of elements,
    which is good if it is a sparse matrix.

    Parameters
    ----------
    array like object : input matrix
    (Qobj or array-like)

    Returns
    -------
    array like object : permutation matrix

    """
    # number of vertices in graph
    n = 0

    # flag for cheking if input is Qobj
    isQobj = False

    # format input for algorithm
    # converts everything into DOK sparse
    if type(A) is list or type(A) is np.ndarray:
        A = dok_matrix(list(A))
        n = A.shape[0]

    elif issparse(A):
        A = dok_matrix(csr_matrix(A))
        n = A.shape[0]

    elif type(A) is qutip.qutip.Qobj:
        A = dok_matrix(csr_matrix(A.data))
        n = A.shape[0]
        isQobj = True
    else:
        raise RuntimeError("Input should be either an "+
            "array-like object or sparse or Qobj")

    # create adjacency list for graph representation
    # of input matrix
    # A is dok_matrix
    temp_dict = list(A.keys())
    graph = {}

    row_with_elems = []
    col_with_elems = []

    for i in temp_dict:
        row_with_elems.append(i[0])
        col_with_elems.append(i[1])
        if(graph.get(i[0]) != None):
            graph[i[0]].append(i[1])
        else:
            graph[i[0]] = []
            graph[i[0]].append(i[1])
        if(graph.get(i[1]) != None):
            graph[i[1]].append(i[0])
        else:
            graph[i[1]] = []
            graph[i[1]].append(i[0])

    row_with_elems = sorted(list(dict.fromkeys(row_with_elems)))
    col_with_elems = sorted(list(dict.fromkeys(col_with_elems)))

    for key in graph.keys():
        graph[key] = sorted(list(dict.fromkeys(graph[key])))
    # print(len(col_with_elems), len(row_with_elems))
    # repeated BFS till all elements have been seen
    seen = [False]*n
    check = False
    block_rows = []
    block_sizes = []

    # total path after all possible disjoint BFS
    net_path = []

    for i in range(n):
        if seen[i] == False:
            seen[i] = True
            # traversal path for single BFS
            # starting from i
            bfs_path = []
            # BFS
            visited = [False]*n
            queue = []
            queue.append(i)
            visited[i] = True
            while queue:
                curr = queue.pop(0)
                bfs_path.append(curr)
                if graph.get(curr) != None:
                    for nbr in graph[curr]:
                        if visited[nbr] == False and seen[nbr] == False:
                            queue.append(nbr)
                            seen[nbr] = True
                            visited[nbr] = True
            block_rows.append(bfs_path)
            block_sizes.append(len(bfs_path))
            net_path.extend(bfs_path)
            # check finds if first BFS covered 
            # the whole graph
            if i == 0:
                if len(bfs_path) < n:
                    check = True
                else:
                    check = False
                    break 

    # checks if connected components are proper subset
    # This means a permutation matrix exists
    if check == True:
        # this means permutation matrix exists
        # computes the permutation matrix
        p_indices = net_path
        rows = range(n)
        vals = [1]*n
        P = csr_matrix((vals,(rows, p_indices)),shape=(n,n))
        return P, block_rows, block_sizes
    else:
        # no permutation matrix exists
        return None, None, None


def get_blocks(A):
    """
    Takes a sparse CSR matrix/numpy array as input 
    and returns number of blocks.

    Algorithm is linear in number of rows, given sparse
    representation. (Needs list of nonzero elements.)

    Parameters
    ----------
    sparse matrix/numpy array: input matrix

    Returns
    -------
    int : number of blocks
    list : all of the blocks themselves
    list : indices at which the blocks start

    """
    
    # total number of rows
    n = A.shape[0]
    # print(A.nonzero()[1])

    # makes dict of indices in each row and col
    row_idx_dict = {row: [] for row in range(n)}
    col_idx_dict = {col: [] for col in range(n)}

    for row, col in zip(*A.nonzero()):
        row_idx_dict[row].append(col)
        col_idx_dict[col].append(row)

    # number of blocks
    block_count = 1

    # for each row maintain a max_before and min_after dict
    max_row_before = {}
    max_col_before = {}
    min_row_after  = {}
    min_col_after  = {}

    # populates before dicts
    max_row = 0
    max_col = 0

    for i in range(n):
        max_row_before[i] = max_row
        max_col_before[i] = max_col
        if(len(row_idx_dict[i]) != 0):
            max_row = max(max_row, row_idx_dict[i][-1])
        else:
            max_row = max_row
        if(len(col_idx_dict[i]) != 0):
            max_col = max(max_col, col_idx_dict[i][-1])
        else:
            max_col = max_col

    # populates after dicts
    min_row = n-1
    min_col = n-1

    for i in reversed(range(n)):
        min_row_after[i] = min_row
        min_col_after[i] = min_col
        if(len(row_idx_dict[i]) != 0):
            min_row = min(min_row, row_idx_dict[i][0])
        else:
            min_row = min_row
        if(len(col_idx_dict[i]) != 0):
            min_col = min(min_col, col_idx_dict[i][0])
        else:
            min_col = min_col

    # stores indices of blocks
    block_indices = []

    # new block begins if 
    # max before is <
    # max after is >=
    for i in range(n):
        if (i > max_row_before[i]   and
            i > max_col_before[i]   and
            i <= min_row_after[i]   and
            i <= min_col_after[i]):
            block_indices.append(i)
            block_count += 1

    # now return as blocks
    block_list = []

    # corner case
    if block_count == 1:
        block_list = None

    else:
        # partitions into actual blocks
        lower_x = 0
        upper_x = n
        for i in range(block_count):

            # set range for finding block
            if i == 0:
                upper_x = block_indices[0]
            elif i == (block_count-1):
                lower_x = upper_x
                upper_x = n
            else:
                lower_x = upper_x
                upper_x = block_indices[i]

            block_list.append(A[lower_x:upper_x, lower_x:upper_x])

    return block_count, block_list, [0]+block_indices


